/* This file is a part of the Amalthea library.
 *
 * Copyright (C) 2018-2021 Eugene 'Vindex' Stulin
 *
 * The Amalthea library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This libary is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see <http://www.gnu.org/licenses/>.
 */

module amalthea.fs;

import core.sys.posix.sys.stat,
       core.sys.posix.dirent,
       core.stdc.errno;
import std.stdio : File;
import std.file, std.path;
public import std.typecons;

public import amalthea.libcore;
import amalthea.sys;
import amalthea.dataprocessing : isAmong;

alias mv  = std.file.rename;
alias cd  = std.file.chdir;
alias pwd = getcwd;
alias rm  = std.file.remove;

import std.algorithm, std.array, std.datetime.systime, std.regex, std.string;
import core.sys.posix.dirent;


/*******************************************************************************
 * Reads entries of the directory.
 * Params:
 *     dir           = Directory for listing files.
 *     hidden        = Whether to include hidden files. Yes, by default.
 *     recursively   = Whether to search in subdirectories. No, by default.
 *     enableFilter  = Whether to filter by type. No, by default.
 *     includedTypes = Types for including to final result.
 *                     Used in conjunction with the previous parameter.
 * See_Also:
 *     $(FS_REF getRegularFiles),
 *     $(FS_REF getDirectories),
 *     $(FS_REF getSymlinks).
 */
FileEntry[] getFiles(string dir,
                     Flag!"hidden" hidden = Yes.hidden,
                     Flag!"recursively" recursively = No.recursively,
                     Flag!"enableFilter" enableFilter = No.enableFilter,
                     FileType[] includedTypes = []) {
    errno = 0;
    DIR* dirstream = opendir(dir.toStringz);
    enforce(!errno, new FileException(getInfoAboutError(errno)));
    dirent* dirEntry = readdir(dirstream);
    enforce(!errno, new FileException(getInfoAboutError(errno)));
    FileEntry[] entries;
    for (; dirEntry; dirEntry = readdir(dirstream)) {
        string name = (*dirEntry).d_name.ptr.fromStringz.idup;
        if (hidden && name.isAmong(".", "..")) continue;
        if (!hidden && name.startsWith('.')) continue;
        FileEntry entry;
        entry.path = getAbsolutePath(chainPath(dir, name).array);
        auto st = FileStat(entry.path);
        entry.fileno = st.inode;
        entry.type = st.type;
        if (recursively && entry.type == FileType.directory) {
            entries ~= getFiles(
                entry.path, hidden, recursively, enableFilter, includedTypes
            );
        }

        if (!enableFilter || entry.type.isAmong(includedTypes)) {
            entries ~= entry;
        }
    }
    enforce(!errno, new FileException(getInfoAboutError(errno)));
    return entries;
}


/*******************************************************************************
 * Gets regular files from the directory.
 * Params:
 *     dir         = Directory for listing regular files.
 *     hidden      = Whether to include hidden files. Yes, by default.
 *     recursively = Whether to search in subdirectories. No, by default.
 */
FileEntry[] getRegularFiles(string dir,
                            Flag!"hidden" hidden = Yes.hidden,
                            Flag!"recursively" recursively = No.recursively) {
    FileType[] includedTypes = [FileType.regularFile];
    return getFiles(dir, hidden, recursively, Yes.enableFilter, includedTypes);
}


/*******************************************************************************
 * Gets directories from the directory.
 * Params:
 *     dir         = Directory for listing directories.
 *     hidden      = Whether to include hidden files. Yes, by default.
 *     recursively = Whether to search in subdirectories. No, by default.
 */
FileEntry[] getDirectories(string dir,
                           Flag!"hidden" hidden = Yes.hidden,
                           Flag!"recursively" recursively = No.recursively) {
    FileType[] includedTypes = [FileType.directory];
    return getFiles(dir, hidden, recursively, Yes.enableFilter, includedTypes);
}


/*******************************************************************************
 * Gets symlinks from the directory.
 * Params:
 *     dir           = Directory for listing symlinks.
 *     hidden        = Whether to include hidden files. Yes, by default.
 *     recursively   = Whether to search in subdirectories. No, by default.
 */
FileEntry[] getSymlinks(string dir,
                        Flag!"hidden" hidden = Yes.hidden,
                        Flag!"recursively" recursively = No.recursively) {
    FileType[] includedTypes = [FileType.symlink];
    return getFiles(dir, hidden, recursively, Yes.enableFilter, includedTypes);
}


/******************************************************************************
 * Special type for files as entries in directories.
 * Analogue of the 'dirent' structure type.
 */
struct FileEntry {
    /// Full path to file.
    string path;
    /// Inode serial number.
    ulong fileno;
    /// UNIX file type. See: enum FileType.
    FileType type;

    /// Returns information about the properties of the file.
    FileStat getFileStat() {
        return FileStat(this.path);
    }
}


/*******************************************************************************
 * Checks if the symbolic link is broken (indicates a non-existent location).
 */
bool isBrokenSymlink(string link)
in {
    assert(link.isSymlink);
}
body {
    return !link.readLink.exists;
}


/*******************************************************************************
 * Special set of functions for work with ini-like .desktop files.
 */
class DesktopEntry {

    /***************************************************************************
     * This function adds new field to ini-like .desktop file.
     * Params:
     *     filepath = path to .desktop file
     *     field = field name
     *     value = value for new field
     */
    static void addField(string filepath, string field, string value) {
        string content = std.file.readText(filepath);
        auto newLine = field ~ "=" ~ value ~ "\n";
        if (content.endsWith("\n")) {
            content ~= newLine;
        } else {
            content ~= "\n" ~ newLine;
        }
        std.file.write(filepath, content);
    }

    /***************************************************************************
     * This function returns value of specified field from .desktop file.
     * Params:
     *     filepath = path to .desktop file
     *     field = field name
     */
    static string getField(string filepath, string field) {
        field = field.replace(`[`, `\[`).replace(`]`, `\]`);
        string content = std.file.readText(filepath);
        auto reg = regex(field ~ `=(.*)`);
        auto res = matchFirst(content, reg);
        return res[1].idup;
    }

    /***************************************************************************
     * This function sets new value for existent field from .desktop file.
     * Params:
     *     filepath = path to .desktop file
     *     field = field name
     *     value = new value of the field
     */
    static void setField(string filepath, string field, string value) {
        string oldValue = DesktopEntry.getField(filepath, field);
        string content = std.file.readText(filepath);
        content = content.replace(field~"="~oldValue, field~"="~value);
        std.file.write(filepath, content);
    }

    /***************************************************************************
     * The function returns title from .desktop file (field "Name").
     */
    static string getTitle(string pathToDesktopFile) {
        string desktopEntry = readText(pathToDesktopFile);
        auto indexOfNameField = desktopEntry.indexOf("Name=");
        auto indexOfTypeField = desktopEntry.indexOf("\nType=");
        return desktopEntry[indexOfNameField + 4 .. indexOfTypeField].idup;
    }

    /***************************************************************************
     * Creates a .desktop file with the link to the network resource.
     * Params:
     *     networkPath = path to the network resource
     *     destinationDir = directory for saving
     *     name = final desktop-file name (if empty, the name is taken from
     *            the title of the HTML page or from network address)
     * Returns: path to created desktop file
     */
    static string createNetworkLink(string networkPath,
                                    string destinationDir,
                                    string name="") {
        import amalthea.net : isLinkToHTML, getHTMLPageTitle;
        if (name.empty) {
            bool HTML = isLinkToHTML(networkPath);
            name = HTML ? getHTMLPageTitle(networkPath)
                        : baseName(networkPath);
            if (name.empty) {
                name = networkPath;
            }
        }
        string desktopEntry = "[Desktop Entry]\n"
                            ~ "Encoding=UTF-8\n"
                            ~ "Name=" ~ name ~ "\n"
                            ~ "Type=Link\n"
                            ~ "URL=" ~ networkPath ~ "\n"
                            ~ "Comment=" ~ "\n"
                            ~ "Icon=text-html";
        string finalFileName = name.replace("/", "|");
        if (!endsWith(name, ".desktop")) {
            finalFileName ~= ".desktop";
        }
        string filePath = chainPath(destinationDir, finalFileName).array;
        std.file.write(filePath, desktopEntry);
        return filePath;
    }

    /***************************************************************************
     * Returns link to network resource from .desktop file (field "URL").
     */
    static string getURL(string pathToDesktopFile) {
        return DesktopEntry.getField(pathToDesktopFile, "URL");
    }
}


/*******************************************************************************
 * Returns absolute path for any raw path
 */
string getAbsolutePath(string path) {
    import std.path : buildNormalizedPath, absolutePath;
    return buildNormalizedPath(absolutePath(path));
}


/*******************************************************************************
 * Function for copying files and directories.
 * Params:
 *     from = source file
 *     to   = destination
 *     followLinks = whether to follow symbolic links
 */
void cp(string from,
        string to,
        Flag!"followLinks" followLinks = No.followLinks) {

    void subfnCopyFile(string src, string dest) {
        if (dest.exists) {
            rm(dest);
        }
        if (!followLinks && src.isSymlink) {
            std.file.symlink(src.readLink, dest);
        } else {
            std.file.copy(src, dest);
        }
    }

    enforce(exists(from), new FileException(from, "Not found."));
    if (from.isDir && to.exists && !to.isDir) {
        throw new std.file.FileException(to, "Is not a directory.");
    }

    if (from.isDir && !to.exists) {
        std.file.mkdirRecurse(to);
    }

    string[] content;
    if (from.isDir && to.isDir) {
        auto startPath = pwd;
        cd(dirName(from));
        scope(exit) cd(startPath);
        content = std.file.dirEntries(from.baseName, SpanMode.depth)
            .map!(a => a.name)
            .array;
        std.file.mkdirRecurse(chainPath(to, baseName(from)).array);
        string[] dirlist = content.filter!(a => a.isDir).array;
        foreach(dir; dirlist) {
            std.file.mkdirRecurse(chainPath(to, dir).array);
        }
        string[] filelist = content.filter!(a => !a.isDir).array;
        foreach(f; filelist) {
            auto dir = std.path.dirName(f);
            subfnCopyFile(f, chainPath(to, dir, baseName(f)).array);
        }
    } else if (!from.isDir && to.exists && to.isDir) {
        subfnCopyFile(from, chainPath(to, baseName(from)).array);
    } else {
        subfnCopyFile(from, to);
    }
}


/*******************************************************************************
 * A structure for storing and retrieving file information
 * on UNIX-like systems.
 */
struct FileStat {
    /// ID of device containing file
    ulong dev;
    /// Inode number
    ulong inode;
    /// File type and mode
    uint mode;
    /// Number of hard links
    ulong nlink;
    /// User ID of owner
    uint uid;
    /// Group ID of owner
    uint gid;
    /// Device ID (if special file)
    ulong rdev;
    /// Total size, in bytes
    ulong size;
    /// Blocksize for file system I/O
    long blocksize;
    /// Number of 512B blocks allocated
    long blocks;
    /// Time of last access
    SysTime atime;
    /// Time of last modification
    SysTime mtime;
    /// Time of last status change
    SysTime ctime;

    /// File path
    string filepath;

    /***************************************************************************
     * The constructor creates object with file information by file path.
     */
    this(string filepath) {
        this.filepath = filepath;
        update();
    }

    /***************************************************************************
     * This method re-reads and updates file information.
     */
    void update() {
        enforce(
            std.file.exists(this.filepath),
            new FileException(this.filepath, "No such file or directory")
        );
        stat_t st;
        lstat(this.filepath.toStringz, &st);
        dev = st.st_dev;
        inode = st.st_ino;
        mode = st.st_mode;
        nlink = st.st_nlink;
        uid = st.st_uid;
        gid = st.st_gid;
        rdev = st.st_rdev;
        size = st.st_size;
        blocksize = st.st_blksize;
        blocks = st.st_blocks;
        atime = SysTime.fromUnixTime(st.st_atime);
        mtime = SysTime.fromUnixTime(st.st_mtime);
        ctime = SysTime.fromUnixTime(st.st_ctime);
    }

    /***************************************************************************
     * Gets user name of owner.
     */
    @property string user() {
        return amalthea.sys.getNameByUID(uid);
    }

    /***************************************************************************
     * Gets group name of owner.
     */
    @property string group() {
        return amalthea.sys.getNameByGID(gid);
    }

    /***************************************************************************
     * Returns file type.
     */
    @property FileType type() {
        switch(this.mode & S_IFMT) {
            case S_IFSOCK: return FileType.socket;
            case S_IFLNK:  return FileType.symlink;
            case S_IFREG:  return FileType.regularFile;
            case S_IFBLK:  return FileType.blockDevice;
            case S_IFDIR:  return FileType.directory;
            case S_IFCHR:  return FileType.charDevice;
            case S_IFIFO:  return FileType.pipe;
            default:       return FileType.unknown;
        }
    }
}


/*******************************************************************************
 * Enumeration of UNIX file types.
 */
enum FileType {
    socket = 's',
    symlink = 'l',
    regularFile = '-',
    blockDevice = 'b',
    directory = 'd',
    charDevice = 'c',
    pipe = 'p',
    unknown = 'z'
}


/*******************************************************************************
 * The function returns file permissions a string form (trwxrwxrwx)
 */
string makeUnixFileModeLine(FileStat st) {
    char[10] permissions;
    permissions[0] = cast(char)st.type;
    foreach(shift, rwxMask; [0: S_IRWXU, 3: S_IRWXG, 6: S_IRWXO]) {
        auto rwxMode = (st.mode & rwxMask) >> (6-shift);
        permissions[shift+1] = (rwxMode & 4) ? 'r' : '-';
        permissions[shift+2] = (rwxMode & 2) ? 'w' : '-';
        permissions[shift+3] = (rwxMode & 1) ? 'x' : '-';
    }
    if (st.mode & S_ISUID) {
        permissions[3] = 's';
    }
    if (st.mode & S_ISGID) {
        permissions[6] = 's';
    }
    if (st.mode & S_ISVTX) {
        permissions[9] = 't';
    }
    return permissions.idup;
}



deprecated("Use std.path.dirName()")
string directoryOfFilePath(string filepath) {
    filepath = filepath.stripRight('/');
    auto lastSlashIndex = filepath.lastIndexOf('/');
    if (lastSlashIndex == -1) return "";
    return filepath[0 .. lastSlashIndex+1].idup;
}


deprecated("Use std.path.baseName()")
string fileNameFromFilePath(string filepath) {
    auto lastSlashIndex = filepath.lastIndexOf('/');
    if (lastSlashIndex == filepath.length-1) return "";
    return filepath[lastSlashIndex+1 .. $].idup;
}



deprecated("Use getFiles()")
string[] getFileList(string catalog,
                     bool baseNames = false,
                     bool hidden = true,
                     bool recursive = false) {
    catalog = getAbsolutePath(catalog);
    string[] filelist, dirlist;
    filelist = std.file.dirEntries(catalog, SpanMode.shallow)
        .filter!(a => a.isFile || a.isSymlink)
        .map!(a => a.name)
        .array;
    dirlist = std.file.dirEntries(catalog, SpanMode.shallow)
        .filter!(a => a.isDir)
        .map!(a => a.name)
        .array;
    if (false == hidden) {
        filelist = filelist.filter!(a => !baseName(a).startsWith(".")).array;
        dirlist  =  dirlist.filter!(a => !baseName(a).startsWith(".")).array;
    }
    if (recursive) foreach(string d; dirlist) {
        string[] temp = getFileList(d, false, hidden, true);
        filelist ~= temp;
    }
    if (baseNames) foreach(ref f; filelist) {
        f = f.baseName;
    }
    return filelist;
}


deprecated("Use getFiles()")
string[] getFileListRecursively(string catalog,
                                bool baseNames = false,
                                bool hidden = true) {
    return getFileList(catalog, baseNames, hidden, true);
}


deprecated("Use getDirectories()")
string[] getDirList(string catalog,
                    bool baseNames = false,
                    bool hidden = true,
                    bool recurse = false) {
    string absoluteCatalog = getAbsolutePath(catalog);
    string[] dirs = std.file.dirEntries(absoluteCatalog, SpanMode.shallow)
        .filter!(a => a.isDir)
        .map!(a => a.name)
        .array;
    if (false == hidden) {
         dirs = dirs.filter!(a => !baseName(a).startsWith(".")).array;
    }
    if (baseNames) dirs = dirs.map!(a => std.path.baseName(a)).array;
    if (recurse) {
        string[] additDirs;
        foreach(dir; dirs) {
            string[] newAdditDirs;

            newAdditDirs = getDirList(
                baseNames ? catalog~"/"~dir : dir,
                baseNames,
                hidden,
                recurse
            );
            if (baseNames) foreach(ref additDir; newAdditDirs) {
                additDir =  dir ~ "/" ~ additDir;
            }
            additDirs ~= newAdditDirs;
        }
        dirs ~= additDirs;
    }
    return dirs;
}


deprecated("Use getDirectories()")
string[] getDirListRecurse(string catalog,
                           bool baseNames = false,
                           bool hidden = true) {
    return getDirList(catalog, baseNames, hidden, true);
}


deprecated("Use getSymlinks()")
string[string] getSymlinksInfo(string catalog,
                               bool showHiddenData = true,
                               bool brokenLinks = false,
                               bool recursively = true) {
    catalog = catalog.stripRight('/');
    string[string] allLinks;
    string[] filelist, dirlist;

    auto files = dirEntries(catalog, SpanMode.shallow);

    if (brokenLinks) {
        foreach(f; files) {
            if (f.isSymlink && (!f.readLink.exists || f.isFile)) {
                filelist ~= baseName(f.name);
            } else if (f.isDir) {
                dirlist ~= baseName(f.name);
            }
        }
    } else {
        foreach(f; files) {
            if (f.isSymlink && !f.readLink.exists) {
                continue;
            } else if (f.isSymlink && f.isFile) {
                filelist ~= baseName(f.name);
            } else if (f.isDir) {
                dirlist ~= baseName(f.name);
            }
        }
    }

    if (showHiddenData == false) {
        filelist = filelist.filter!(a => !a.startsWith(".")).array;
        dirlist  =  dirlist.filter!(a => !a.startsWith(".")).array;
    }
    foreach(string link; filelist) {
        allLinks[absolutePath(catalog)~"/"~link] = readLink(catalog~'/'~link);
    }
    if (recursively)
        foreach(string d; dirlist) {
            string[string] temp = getSymlinksInfo(catalog~'/'~d,
                                                  showHiddenData,
                                                  brokenLinks,
                                                  recursively);
            foreach(pathToLink, pathToFile; temp) {
                allLinks[pathToLink] = pathToFile;
            }
        }
    return allLinks;
}
deprecated alias getFileLinksInfo = getSymlinksInfo;


deprecated("Use std.file.getTimes or amalthea.fs.FileStat")
SysTime timeLastAccessed(string name) {
    SysTime accessTime, modTime;
    std.file.getTimes(name, accessTime, modTime);
    return accessTime;
}


deprecated("Use std.file.getTimes or amalthea.fs.FileStat")
SysTime timeLastModified(string name) {
    SysTime accessTime, modTime;
    std.file.getTimes(name, accessTime, modTime);
    return modTime;
}


deprecated("Use std.stdio.tmpfile()")
string createTempFile() {
    string tempfile;
    do {
        tempfile = std.file.tempDir ~ "/tempfile"
                 ~ Clock.currTime.toISOExtString();
    } while(std.file.exists(tempfile));
    std.file.write(tempfile, "");
    return tempfile;
}
deprecated alias getTempFilePath = createTempFile;

